#include <math.h>
#include <stdio.h>
#include <string.h>



#include "NotesClasses.hpp"

//=============================================================================
//=============================================================================
//    I N T E R P O L A T E D    M O D E L    C L A S S
//=============================================================================
//=============================================================================
int InterpolatedModel::MakePureTone()
{
//--- Set Samples

        double theRadiansPerSample = 2.0 * M_PI / (double) NOTES_MODEL_SIZE;

        for( int s = 0; s < NOTES_MODEL_SIZE; s++ )
        {
            MySamples[s] = sin( theRadiansPerSample * s )
                         * 1 * ( 32 * 1024 - 1 );
        }

//--- Set Deltas as circular

        SetDeltas();

        return 0;
}
//=============================================================================
int InterpolatedModel::MakeADSR()
{
        int theIndex = 0;
        int theSlope = 4 * 1024 * 4 / NOTES_MODEL_SIZE;
    
//--- Attack

        for( int s = 0; s < NOTES_MODEL_SIZE/4; s++ )
        {
            MySamples[theIndex] = s * theSlope;
            theIndex++;
        }

//--- Decay

        theSlope /= 2;

        for( int s = 0; s < NOTES_MODEL_SIZE/4; s++ )
        {
            MySamples[theIndex] = 4 * 1024 - s * theSlope - 1;
            theIndex++;
        }

//--- Release

        for( int s = 0; s < NOTES_MODEL_SIZE/2; s++ )
        {
            MySamples[theIndex] = 253 * MySamples[theIndex-1] / 256;
            theIndex++;
        }

//--- Set Deltas as circular

        SetDeltas();

        return 0;

}
//=============================================================================
int InterpolatedModel::SetDeltas()
{
        for( int s = 0; s < NOTES_MODEL_SIZE-1; s++ )
        {
            MyDeltas[s] = MySamples[s+1] - MySamples[s];
        }

        MyDeltas[NOTES_MODEL_SIZE-1] = MySamples[0]
                                     - MySamples[NOTES_MODEL_SIZE-1];

        return 0;
}
//=============================================================================
int InterpolatedModel::ValueOf( double ArgSpot )
{
        int theIndex = (int) floor( ArgSpot );

        if( theIndex >= NOTES_MODEL_SIZE ) return 0;

        int theNotch = (int) ( ( ArgSpot - theIndex ) * 1024 );

        int theTweak = ( theNotch * MyDeltas[theIndex] ) >> 10;

        return MySamples[theIndex] + theTweak;
}
//=============================================================================





//=============================================================================
//=============================================================================
//    P L A Y     C U R S O R
//=============================================================================
//=============================================================================
int PlayCursor::Initialize( double ArgFrequencyHz, double ArgAttackSec )
{
        MyWaveFormStep = NOTES_MODEL_SIZE * ArgFrequencyHz / 44100.0;
        MyProfileStep  = ( NOTES_MODEL_SIZE/4 ) / ( ArgAttackSec * 44100.0 );
    
        MyWaveFormSpot = 0.0;
        MyProfileSpot  = 0.0;

        MyState = NOTES_STATE_ATTACK;

        return 0;
}
//=============================================================================
int PlayCursor::Increment( double ArgDecaySec,
                           double ArgReleaseSec,
                           int    ArgReleasedFlag )
{
//--- Along the waveform

        MyWaveFormSpot += MyWaveFormStep;

        if( MyWaveFormSpot >= NOTES_MODEL_SIZE )
        {
            MyWaveFormSpot -= NOTES_MODEL_SIZE;
        }

//--- Along the Profile

        MyProfileSpot += MyProfileStep;

//printf( "MyProfileSpot = %f\n", MyProfileSpot );    


//--- Look for state change

        int theProfileIndex = (int) floor( MyProfileSpot );

        switch( MyState )
        {
          case NOTES_STATE_ATTACK:
            if( theProfileIndex >= NOTES_MODEL_SIZE/4 )
            {
                MyState       =   NOTES_STATE_DECAY;
                MyProfileStep = ( NOTES_MODEL_SIZE/4 )
                              / ( ArgDecaySec * 44100.0 );
            }
            break;

          case NOTES_STATE_DECAY:
            if( theProfileIndex >= NOTES_MODEL_SIZE/2 )
            {
                if( ArgReleasedFlag )
                {
                    MyState       =   NOTES_STATE_RELEASE;
                    MyProfileStep = ( NOTES_MODEL_SIZE/2 )
                                  / ( ArgReleaseSec * 44100.0 );

                }
                else
                {
                    MyState       = NOTES_STATE_SUSTAIN;
                    MyProfileStep = 0.0;
                }
            }
            break;

          case NOTES_STATE_SUSTAIN:
            if( ArgReleasedFlag )
            {
                MyState        =   NOTES_STATE_RELEASE;
                MyProfileStep  = ( NOTES_MODEL_SIZE/2 )
                               / ( ArgReleaseSec * 44100.0 );
            }
            break;

          case NOTES_STATE_RELEASE:
            if( theProfileIndex >= NOTES_MODEL_SIZE )
            {
                MyState = NOTES_STATE_EXPIRED;
                MyProfileStep = 0.0;
                MyProfileSpot = NOTES_MODEL_SIZE - 1.0;
            }
            break;
        }

        return 0;
    
}
//=============================================================================





//=============================================================================
//=============================================================================
//    I N S T R U M E N T    P L A C E    C L A S S
//=============================================================================
//=============================================================================
int InstrumentPlace::Define( InterpolatedModel* ArgWaveForm,
                             InterpolatedModel* ArgProfile,
                             double             ArgAttackSec,
                             double             ArgDecaySec,
                             double             ArgReleaseSec,
                             double             ArgVolumePan,
                             double             ArgDelayPanSec )
{
//--- Store passed values

        MyWaveForm = ArgWaveForm;
        MyProfile  = ArgProfile;

        MyAttackSec   = ArgAttackSec;
        MyDecaySec    = ArgDecaySec;
        MyReleaseSec  = ArgReleaseSec;

        return 0;
}
//=============================================================================
int InstrumentPlace::GetProfileTimes( double  ArgFrequencyHz,
                                      double* RetAttackSec,
                                      double* RetDecaySec,
                                      double* RetReleaseSec )
{                             
        *RetAttackSec  = MyAttackSec;
        *RetDecaySec   = MyDecaySec;
        *RetReleaseSec = MyReleaseSec;

        return 0;
}
//=============================================================================
int InstrumentPlace::GetSampleValue( PlayCursor* ArgCursor )
{                             
        int theWaveFormValue = MyWaveForm->ValueOf( ArgCursor->MyWaveFormSpot );

        int theProfileValue  = MyProfile->ValueOf(  ArgCursor->MyProfileSpot );

        return ( theWaveFormValue * theProfileValue ) >> 6;
}
//=============================================================================





//=============================================================================
//=============================================================================
//    P L A Y I N G    N O T E    C L A S S
//=============================================================================
//=============================================================================
int PlayingNote::Activate( int                ArgChunkNumber,
                           InstrumentPlace*   ArgInstrument,
                           double             ArgFrequencyHz )
{
//--- Store passed values

        MyStartChunk  = ArgChunkNumber;

        MyInstrument  = ArgInstrument;
        MyFrequencyHz = ArgFrequencyHz;

//--- Get the profile times

        ArgInstrument->GetProfileTimes( ArgFrequencyHz,
                                        &MyAttackSec,
                                        &MyDecaySec,
                                        &MyReleaseSec );

//--- Initialize tracking parameters       

        MyReleasedFlag = 0;

        MyLeftCursor.Initialize(  MyFrequencyHz, MyAttackSec );
        MyRightCursor.Initialize( MyFrequencyHz, MyAttackSec );

        return 0;
}
//=============================================================================
int PlayingNote::Release()
{
        MyReleasedFlag = 1;
        return 0;
}
//=============================================================================
int PlayingNote::AddToChunk( int* ArgChunkData )
{
        addToChannel( &ArgChunkData[0], &MyLeftCursor  );
        addToChannel( &ArgChunkData[1], &MyRightCursor );

        return 0;
}
//=============================================================================
int PlayingNote::addToChannel( int*        argChannelData,
                               PlayCursor* argCursor )
{
        int* p = argChannelData;

        for( int s = 0; s < 441; s++ )
        {
            int theValue = MyInstrument->GetSampleValue( argCursor );

            *p += theValue;  p += 2;

            argCursor->Increment( MyDecaySec, MyReleaseSec, MyReleasedFlag );
        }

        return 0;
}
//=============================================================================





//=============================================================================
//=============================================================================
//    S T R E A M    J O C K E Y    C L A S S
//=============================================================================
//=============================================================================
int StreamJockey::Start()
{
//--- Allocate working space

        MyWaveForms    = new InterpolatedModel[100];
        MyProfiles     = new InterpolatedModel[100];

        MyInstruments  = new InstrumentPlace[100];

        MyPlayingNotes = new PlayingNote[1000];

        MyChunkData    = new int[882];
        MyChunkStream  = new short int[882];

//--- Define standard tones

        MyWaveForms[0].MakePureTone();

//--- Define standard profiles

        MyProfiles[0].MakeADSR();

//--- Create a default instrument

        MyInstruments[0].Define( &MyWaveForms[0],
                                 &MyProfiles[0],
                                 0.1, 0.1, 0.5,
                                 0.0, 0.0 );

        MyCurrentInstrument = &MyInstruments[0];

//--- Set PulseAudio parameters

        MySpecifications.format   = PA_SAMPLE_S16LE;
        MySpecifications.rate     = 44100;
        MySpecifications.channels = 2;

//--- Set playing notes pointers

        for( int pn = 0; pn < 999; pn++ )
        {
            MyPlayingNotes[pn].MyNextNote = &MyPlayingNotes[pn+1];
        }

        MyPlayingNotes[999].MyNextNote = NULL;

        MyFirstAvailableNote = &MyPlayingNotes[0];
        MyFirstBusyNote      = NULL;

//--- Create a playback stream

        int theError;

        MyStream = pa_simple_new( NULL,
                                  "libNotes",
                                  PA_STREAM_PLAYBACK,
                                  NULL,
                                  "playback",
                                  &MySpecifications,
                                  NULL,
                                  NULL,
                                  &theError );

        if( !MyStream )
        {
            printf( "pa_simple_new() failed: %s\n",
                    pa_strerror( theError ) );

            return -1;
        }

        MyChunkNumber = 0;

//--- Exit

        return 0;
}
//=============================================================================
int StreamJockey::Finish()
{
//--- Drain

        int theError;

        if( pa_simple_drain( MyStream, &theError ) < 0 )
        {
            printf( "pa_simple_drain() failed: %s\n",
                    pa_strerror( theError ) );
        }

//--- Free the stream

        if( MyStream ) pa_simple_free( MyStream );

//--- Free the working space

        delete MyWaveForms;
        delete MyProfiles;

        delete MyInstruments;

        delete MyChunkData;

        return 0;
}
//=============================================================================
PlayingNote* StreamJockey::AddNote( double ArgFrequencyHz )
{
//--- Obtain available note

        PlayingNote* theNote = MyFirstAvailableNote;

        if( theNote == NULL ) return NULL;

//--- Remove from available list and add to busy list

        MyFirstAvailableNote = theNote->MyNextNote;
        theNote->MyNextNote  = MyFirstBusyNote;
        MyFirstBusyNote      = theNote;

//--- Activate the note

        theNote->Activate( MyChunkNumber,
                           MyCurrentInstrument,
                           ArgFrequencyHz );
 
        return theNote;
}
//=============================================================================
int StreamJockey::Release( PlayingNote* ArgNote )
{
        ArgNote->Release();
        return 0;
}
//=============================================================================
int StreamJockey::Tick()
{
//--- Zero the chunk data buffer

        memset( MyChunkData, '\0', 4 * 882 );

//--- Submit current playing notes

        PlayingNote* theChunkNote = MyFirstBusyNote;
        PlayingNote* thePriorNote = NULL;

        while( theChunkNote != NULL )
        {
            theChunkNote->AddToChunk( MyChunkData );

            PlayingNote* theNextNote = theChunkNote->MyNextNote; 

            if( theChunkNote->MyLeftCursor.MyState  == NOTES_STATE_EXPIRED &&
                theChunkNote->MyRightCursor.MyState == NOTES_STATE_EXPIRED )
            {
                theChunkNote->MyLeftCursor.MyState  = NOTES_STATE_IDLE;
                theChunkNote->MyRightCursor.MyState = NOTES_STATE_IDLE;
                
                if( thePriorNote == NULL )
                {
                    MyFirstBusyNote = theNextNote;
                }
                else
                {
                    thePriorNote->MyNextNote = theNextNote;
                }

                theChunkNote->MyNextNote = MyFirstAvailableNote;
                MyFirstAvailableNote     = theChunkNote;
            }
            else
            {
                thePriorNote = theChunkNote;
            }

            theChunkNote = theNextNote;
        }

//--- Limiter

        int theExcess;

        for( int s = 0; s < 882; s++ )
        {
            int theValue = MyChunkData[s] >> 8;

            if( theValue < -30000 )
            {
                theExcess = -MyChunkStream[s] - 30000;
                theValue  = -32767 + 2767 * 2767 / ( 2767 + theExcess );
            }
            else if( theValue > 30000 )
            {
                theExcess = MyChunkStream[s] - 30000;
                theValue  = 32767 - 2767 * 2767 / ( 2767 + theExcess );
            }

            MyChunkStream[s] = (short int) theValue;
        }

//--- Submit to playback

        int theError;

        int theResult = pa_simple_write( MyStream,
                                         MyChunkStream,
                                         2 * 2 * 441,
                                         &theError );

        if( theResult < 0)
        {
            printf( "pa_simple_write() failed: %s\n",
                    pa_strerror( theError ) );
        }

//--- Increment chunk number

        MyChunkNumber++;

        return 0;
}
//=============================================================================
int StreamJockey::GetRemaining()
{
        return 0; //theAvailable;
}
//=============================================================================



