#include <pulse/simple.h>
#include <pulse/error.h>

#define NOTES_MODEL_SIZE    4096

#define NOTES_STATE_IDLE       0
#define NOTES_STATE_HUSH       1
#define NOTES_STATE_ATTACK     2
#define NOTES_STATE_DECAY      3
#define NOTES_STATE_SUSTAIN    4
#define NOTES_STATE_RELEASE    5
#define NOTES_STATE_QUIETED    6
#define NOTES_STATE_EXPIRED    7

#define NOTES_BOARD_ACROSS     1
#define NOTES_BOARD_SPLIT      2


//-----------------------------------------------------------------------------
class InterpolatedModel
{
public:

    int     MySamples[NOTES_MODEL_SIZE];
    int     MyDeltas[NOTES_MODEL_SIZE];

//--- Methods

    int MakePureTone();
    int MakeADSR();
    int SetDeltas();
    int ValueOf( double ArgSpot );
};
//-----------------------------------------------------------------------------
class PlayCursor
{
public:

    double              MyWaveFormSpot;
    double              MyWaveFormStep;

    double              MyProfileSpot;
    double              MyProfileStep;

    int                 MyState;

//--- Methods

    int Initialize(  double ArgFrequencyHz, double ArgAttackSec );

    int Increment( double ArgDecaySec,
                   double ArgReleaseSec,
                   int    ArgReleasedFlag );


};
//-----------------------------------------------------------------------------
class InstrumentPlace
{
public:

    InterpolatedModel*  MyWaveForm;
    InterpolatedModel*  MyProfile;

    double              MyAttackSec;
    double              MyDecaySec;
    double              MyReleaseSec;

//--- Methods

    int              Define( InterpolatedModel* ArgWaveForm,
                             InterpolatedModel* ArgProfile,
                             double             ArgAttackSec,
                             double             ArgDecaySec,
                             double             ArgReleaseSec,
                             double             ArgVolumePan,
                             double             ArgDelayPanSec );

    int              GetProfileTimes( double  ArgFrequencyHz,
                                      double* RetAttackSec,
                                      double* RetDecaySec,
                                      double* RetReleaseSec );

    int              GetSampleValue( PlayCursor* ArgCursor );

};
//-----------------------------------------------------------------------------
class PlayingNote
{
public:

    double              MyFrequencyHz;

    double              MyAttackSec;
    double              MyDecaySec;
    double              MyReleaseSec;

    int                 MyStartChunk;
    int                 MyReleasedFlag;

    PlayCursor          MyLeftCursor;
    PlayCursor          MyRightCursor;

    InstrumentPlace*    MyInstrument;

    PlayingNote*        MyNextNote;
   
//--- Methods
    
    int         Activate( int                ArgChunkNumber,
                          InstrumentPlace*   ArgInstrument,
                          double             ArgFrequencyHz );
                           
    int         Release();

    int         AddToChunk( int* ArgChunkData );


private:

    int         addToChannel( int*        argChannelData,
                              PlayCursor* argCursor );

};
//-----------------------------------------------------------------------------
class StreamJockey
{
public:

    pa_sample_spec      MySpecifications;
    pa_simple*          MyStream;
    
    InterpolatedModel*  MyWaveForms;
    InterpolatedModel*  MyProfiles;

    InstrumentPlace*    MyInstruments;
    InstrumentPlace*    MyCurrentInstrument;

    int*                MyChunkData;
    short int*          MyChunkStream;

    PlayingNote*        MyPlayingNotes;
    PlayingNote*        MyFirstAvailableNote;
    PlayingNote*        MyFirstBusyNote;

    unsigned int        MyChunkNumber;

//--- Methods

    int Start();
    int Finish();

    PlayingNote* AddNote( double ArgFrequencyHz );
    int          Release( PlayingNote* ArgNote );

    int Tick();

    int GetRemaining();

};
//-----------------------------------------------------------------------------

