#include <math.h>
#include <stdio.h>
#include <string.h>

#include "NotesClasses.hpp"

//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
extern "C"    //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
{             //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    StreamJockey MyStreamJockey;

    int          MyKeyMidi[256];
    double       MyKeyFrequency[256];
    PlayingNote* MyKeyNote[256];

    double       MyMidiFrequency[128];
    PlayingNote* MyMidiNote[128];


//=============================================================================
//=============================================================================
//    L I B R A R Y    H O U S E K E E P I N G
//=============================================================================
//=============================================================================
__attribute__((constructor)) void init(void)
{

}
//=============================================================================
__attribute__((destructor)) void fini(void)
{

}
//=============================================================================
//=============================================================================
//    L I B R A R Y    P R I V A T E    U T I L I T I E S
//=============================================================================
//=============================================================================
void notes_SetFrequencies( int argMidiNumber, const char* argKeys )
{
        const char* p = argKeys;

        int m = argMidiNumber;

        for(;;)
        {
            int theKey = (int) *p;
            if( theKey == 0 ) break;

            MyKeyFrequency[theKey] = MyMidiFrequency[m];
            MyKeyMidi[theKey]      = m;
            
            m++;
            p++;
        }
}
//=============================================================================
void notes_AssignKeyboardLayout( int argBoardType )
{
        switch( argBoardType )
        {
          case NOTES_BOARD_ACROSS:
                                  //   C D_EF_G A B    
            notes_SetFrequencies( 48, "1234567890-=" );     // 3
            notes_SetFrequencies( 60, "QWERTYUIOP[]" );     // 4
            notes_SetFrequencies( 72, "ASDFGHJKL;'\n" );    // 5 
            notes_SetFrequencies( 84, "ZXCVBNM,./" );       // 6 
            break;

          case NOTES_BOARD_SPLIT:
                                  //   C D_EF
                                  //   _G A B    
            notes_SetFrequencies( 48, "123456" );     // 3
            notes_SetFrequencies( 54, "QWERTY" );     // 4
            notes_SetFrequencies( 60, "ASDFGH" );     // 5
            notes_SetFrequencies( 66, "ZXCVBN" );     // 6

            notes_SetFrequencies( 72, "7890-=" );     // 3
            notes_SetFrequencies( 78, "UIOP[]" );     // 4
            notes_SetFrequencies( 84, "JKL;'\n" );    // 5 
            notes_SetFrequencies( 90, "M,./" );       // 6 
            break;
        }

    
}
//=============================================================================
//=============================================================================
//    L I B R A R Y    E N T R Y    P O I N T S
//=============================================================================
//=============================================================================
int Notes_Start()
{
//--- Initialize key to note lookup table

        for( int k = 0; k < 256; k++ )
        {
            MyKeyFrequency[k] = -1.0;
            MyKeyNote[k]      =  NULL;
        }

//--- Set the MIDI frequencies

        for( int m = 0; m < 128; m++ )
        {
            MyMidiFrequency[m] = pow( 2.0, ( m - 69.0 ) / 12.0 ) * 440.0;
            MyMidiNote[m]      = NULL;
        }

//--- Assign keyboard notes

        notes_AssignKeyboardLayout( NOTES_BOARD_ACROSS );

//--- Start the jockey

        MyStreamJockey.Start();

        return 0;
}
//=============================================================================
int Notes_Finish()
{
        MyStreamJockey.Finish();

        return 0;
}
//=============================================================================
int Notes_KeyPress( int ArgKeyCode )
{
//--- Handle Special keycodes

        int theKey = ArgKeyCode & 0xFF;

        if( ( ArgKeyCode & 0xFFFFFF00 ) > 0 )
        {
            switch( ArgKeyCode )
            {
              case 0x01000004:
                theKey = 10;
                break;

              default:
                return -1;
            }
        }

//--- Bail on non note keys

        if( MyKeyFrequency[theKey] <= 0.0 ) return -2;

//--- Add note if not repeat key

        if( MyKeyNote[theKey] == NULL )
        {
            MyKeyNote[theKey] = MyStreamJockey.AddNote( MyKeyFrequency[theKey] );
        }

        return 0;
}
//=============================================================================
int Notes_KeyRelease( int ArgKeyCode )
{
//--- Handle Special keycodes

        int theKey = ArgKeyCode & 0xFF;

        if( ( ArgKeyCode & 0xFFFFFF00 ) > 0 )
        {
            switch( ArgKeyCode )
            {
              case 0x01000004:
                theKey = 10;
                break;

              default:
                return -1;
            }
        }

//--- Find the current note

        PlayingNote* theNote = MyKeyNote[theKey];

        if( theNote == NULL ) return -2;

//--- Clear the Key state

        MyStreamJockey.Release( theNote );

        MyKeyNote[theKey] = NULL;

        return 0;
}
//=============================================================================
int Notes_MidiPress( int ArgMidiCode )
{
//--- Add note if not repeat key

        if( MyMidiNote[ArgMidiCode] == NULL )
        {
            MyMidiNote[ArgMidiCode] = MyStreamJockey.AddNote( MyMidiFrequency[ArgMidiCode] );
        }

        return 0;
}
//=============================================================================
int Notes_MidiRelease( int ArgMidiCode )
{
//--- Find the current note

        PlayingNote* theNote = MyMidiNote[ArgMidiCode];

        if( theNote == NULL ) return -2;

//--- Clear the Key state

        MyStreamJockey.Release( theNote );

        MyMidiNote[ArgMidiCode] = NULL;

        return 0;
}
//=============================================================================
int Notes_Tick()
{
        MyStreamJockey.Tick();

        return 0;
}
//=============================================================================


//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
}             //:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// extern "C"   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
